import logging
from path import Path

import appdirs
import pydash as py_
from addict import Dict

from easytrajh5.fs import dump_yaml, load_yaml

logger = logging.getLogger(__name__)

DEFAULT = Dict(
    chimera="/Applications/Chimera.app/Contents/MacOS/chimera",
    pymol="pymol",
    vmd="/Applications/VMD\ 1.9.4a51-arm64-Rev9.app/Contents/vmd/vmd_MACOSXARM64",
)

binary_yaml = appdirs.user_data_dir("rseed.binary.yaml")

if Path(binary_yaml).exists():
    logger.info(f"user configs {binary_yaml}")
    binary = load_yaml(binary_yaml)
else:
    binary = py_.clone_deep(DEFAULT)
    logger.info(f"create binary config file {binary_yaml}")
    dump_yaml(binary, binary_yaml)


def get_binary(key, test_binary=None):
    if test_binary is not None:
        if Path(test_binary).exists():
            return test_binary
    result = binary.get(key)
    print(f"Binaries config `{binary_yaml}`")
    print(f"Entry for {key}: `{result}`")
    return result



import logging

import mdtraj
import numpy
import parmed
import pydash as py_
from parmed.amber import AmberMask
from path import Path

from .fs import load_yaml
from .struct import (
    get_mdtraj_from_parmed,
    calc_residue_contacts_with_mdtraj,
    slice_parmed,
)

logger = logging.getLogger(__name__)
data_dir = Path(__file__).parent / "data"
resnames_by_keyword = load_yaml(data_dir / "select.yaml")


def select_mask(pmd, mask, is_fail_on_empty=True):
    """
    Selects atom based on a selection string which is based on a combination
    atom selection langauge.There are several different types of selection
    modes strings, where the first word is often used a mode selector.

    - keywords
        - accepts (in any order): 'ligand', 'protein', 'water', 'lipid', 'salt',
          'solvent', 'lipid', 'nucleic', 'resname', 'resi', 'atom'
        - If more than one keyword is specified, it is assumed they are joined with "or"
          operation (i.e. 'ligand protein' will return both ligand and protein atom indices).
        - 'ligand' will find the residue 'LIG', 'UNL', 'UNK', or
          whatever is in 'ligand' in the h5 'easytrajh5/data/select.yaml'
        - 'pocket' will find the closest 6 residues to the 'ligand' group.
        - 'near' will require a following resname, with an optional integer, e.g.:
            'near ATP'
            'near ATP 5'
        - 'resname' identifies a single residue type (usually a ligand):
            'resname LEU'
        - 'resi' for residue 0-indexed selections
            "resi 0 10-13" - selects atoms in the first and 11th, 12th and 13th residues
        - 'atom' atom 0-indexed selections
            "atom 0 55 43 101-105" - selects the first, 56th, 44th, 101 to 105th atom
    - AMBER-style atom selection https://parmed.github.io/ParmEd/html/amber.html#amber-mask-syntax
        "amber :ALA,LYS" - selects all alanine and lysine residues
    - MDTraj-style atom selection - https://mdtraj.org/1.9.4/atom_selection.html
        "mdtraj protein and water" - selects protein and water
    - furthermore, selections can be combined with set operators ("not", "intersect", "merge", "diff"),
        "intersect {not {amber :ALA}} {protein}"
        "diff {protein} {not {amber :ALA}}"
        "not {resname LEU}"
        "merge {near BSM 8} {amber :ALA}"
    - some useful masks:
        - no solvent "not {solvent}"
        - just the protein "protein"
        - heavy protein atoms "diff {protein} {amber @/H}"
        - no hydrogens "not {amber @/H}"
        - pocket and ligand "pocket ligand"
        - specified ligand with 10 closest neighbours "resname UNL near UNL 10"

    :parmam parmed_structure: parmed.Structure
    :param mask: Union[str, List[int]]

    :return: [int] - sorted list of atom indices to pmd
    """

    def get_i_atoms_of_ast(ast):
        has_list = py_.some(ast, lambda node: isinstance(node, list))
        if not has_list:
            expr = " ".join(ast)
            value = process_expr(pmd, expr)
            return value

        operator = ast[0]

        if operator == "not":
            if len(ast) != 2:
                raise ValueError("not: must have only 1 following expr in {}")
            i_atoms1 = get_i_atoms_of_ast(ast[1])
            i_all_atoms = [a.idx for a in pmd.atoms]
            return diff_list(i_all_atoms, i_atoms1)

        elif operator == "merge":
            result = []
            for node in ast[1:]:
                result.extend(get_i_atoms_of_ast(node))
            return list(set(result))

        elif operator == "intersect":
            result = set(get_i_atoms_of_ast(ast[1]))
            for node in ast[2:]:
                i_atom_set = set(get_i_atoms_of_ast(node))
                result = result.intersection(i_atom_set)
            return list(result)

        elif operator == "diff":
            if len(ast) != 3:
                raise ValueError("diff: must have 2 expr in {}")
            i_atoms1 = get_i_atoms_of_ast(ast[1])
            i_atoms2 = get_i_atoms_of_ast(ast[2])
            return diff_list(i_atoms1, i_atoms2)

        else:
            raise ValueError(f"Operator {operator} not in [not diff merge intersect]")

    if isinstance(mask, str):
        i_atoms = get_i_atoms_of_ast(parse_ast(mask))
    elif is_integer_list(mask):
        i_atoms = mask
    else:
        raise ValueError(f"Can't parse atom mask {mask}")

    if is_fail_on_empty and not len(i_atoms):
        raise ValueError("Selection produced no atoms")

    i_atoms = py_.uniq(i_atoms)
    i_atoms.sort()

    res_indices = py_.uniq([pmd.atoms[i].residue.idx for i in i_atoms])
    logger.info(
        f'select_mask "{mask}" -> {len(i_atoms)} atoms, {len(res_indices)} residues'
    )

    return i_atoms


def is_integer_list(o):
    if not isinstance(o, (list, numpy.ndarray)):
        return False
    for x in o:
        if not isinstance(x, (int, numpy.integer)):
            return False
    return True


def process_expr(pmd, expr):
    expr_lower = expr.lower()
    if expr_lower.startswith("amber "):
        amber_mask = AmberMask(pmd, expr[6:])
        return [i_atom for i_atom, mask in enumerate(amber_mask.Selection()) if mask]
    elif expr_lower.startswith("mdtraj "):
        mdtraj_top = mdtraj.Topology.from_openmm(pmd.topology)
        return mdtraj_top.select(expr[6:]).tolist()
    else:
        return select_keywords(pmd, expr)


def parse_ast(mask):
    """
    Parses an expression bounded by curly brackets into an abstract syntax tree
    e.g. 'a {{b c} {d}} e}' -> ['a', [[['b', 'c'], ['d'], 'e']]
    """

    def push(obj, a_list, depth):
        while depth:
            a_list = a_list[-1]
            depth -= 1
        a_list.append(obj)

    def parse_parentheses(tokens):
        groups = []
        depth = 0

        try:
            for token in tokens:
                if token == "{":
                    push([], groups, depth)
                    depth += 1
                elif token == "}":
                    depth -= 1
                else:
                    push(token, groups, depth)
        except IndexError:
            raise ValueError("Parentheses mismatch")

        if depth > 0:
            raise ValueError("Parentheses mismatch")
        else:
            return groups

    mask = mask.replace("{", " { ").replace("}", " } ")
    tokens = py_.filter_(mask.split(" "))
    return parse_parentheses(tokens)


def parse_number_list(num_str_or_list):
    """
    Turns a string "1,2,3-4,5, 7 9-11 8-33" into a list of numbers
    """
    if isinstance(num_str_or_list, str):
        result = []
        parse_s = num_str_or_list.replace("-", " - ").replace(",", " ")
        tokens = parse_s.split()
        n = len(tokens)
        start = None
        for i in range(n):
            if tokens[i] == "-":
                continue
            num = int(tokens[i])
            if i < n - 1 and tokens[i + 1] == "-":
                start = num
                continue
            if start is not None:
                for i in range(start, num + 1):
                    result.append(i)
                start = None
            else:
                result.append(num)
        return result
    elif is_integer_list(num_str_or_list):
        return num_str_or_list
    else:
        raise ValueError(f"Can't process {num_str_or_list}")


def select_resi(pmd, expr):
    i_residues = parse_number_list(expr)
    result = []
    for residue in pmd.topology.residues():
        if residue.index in i_residues:
            for a in residue.atoms():
                result.append(a.index)
    return result


def select_residue_contacts(pmd, lig_resnames=["LIG"], max_n_residue=6, cutoff_nm=None):
    """
    Finds the atoms of the residues closest to a ligand residue

    :return: [int]
    """
    logger.info(f"select_residue_contacts to {lig_resnames}")

    traj = get_mdtraj_from_parmed(pmd)
    residues = list(traj.topology.residues)
    protein_resnames = get_resnames("protein")
    i_ligand_residues = [r.index for r in residues if r.name in lig_resnames]
    i_protein_residues = [r.index for r in residues if r.name in protein_resnames]

    if len(i_ligand_residues) == 0:
        raise ValueError(f"no residue with resname={lig_resnames}")

    i_closest_residues = calc_residue_contacts_with_mdtraj(
        traj,
        i_ligand_residues,
        i_protein_residues,
        cutoff_nm=cutoff_nm,
        max_n_residue=max_n_residue,
    )

    logger.info(f"found {len(i_closest_residues)} residues")

    return select_resi(pmd, i_closest_residues)


def get_resnames(keyword):
    return resnames_by_keyword[keyword]


def select_resnames(pmd, resnames):
    return [a.idx for a in pmd.atoms if a.residue.name in resnames]


def diff_list(list1, list2):
    return list(sorted(set(list1) - set(list2)))


def select_keywords(pmd: parmed.Structure, expr: str) -> [int]:
    """
    Select a list of atom  that can be used to slice parmed objects, set restraints, or specify CVs.
    If both keyword and resid selection are applied both are returned (i.e. 'or' combining is assumed).
        - accepts (in any order): 'ligand', 'protein', 'water', 'lipid', 'salt', 'solvent', 'lipid', 'nucleic'
          If more than one keyword is specified, it is assumed they are joined with "or"
          operation (i.e. 'ligand protein' will return both ligand and protein atom indices).
        - The keyword 'ligand' will find the residue 'LIG', 'UNL', 'UNK', or
          whatever is in 'ligand_resname'
        - The keyword 'pocket' will find the closest 6 residues to the ligand.
        - The keyword 'near' will require a following resname, with an optional integer, e.g.:
            'near ATP'
            'near ATP 5'
        - The keyword 'resname' will require a resname:
            'resname LEU'

    :param expr:
    """
    result = []
    if expr:
        tokens = [t for t in expr.split(" ") if t]
        keywords = list(resnames_by_keyword.keys())
        deprecated = [
            "noh",
            "nosolvent",
            "resid",
        ]
        operants = [
            "atom",
            "resi",
            "pocket",
            "near",
            "resname",
        ]
        allowed_keywords = keywords + operants + deprecated
        while len(tokens) > 0:
            keyword = tokens.pop(0).lower()
            if keyword not in allowed_keywords:
                raise ValueError(f"Keyword {keyword} not in {allowed_keywords}")
            i_atoms = []
            if keyword in resnames_by_keyword:
                i_atoms = select_resnames(pmd, get_resnames(keyword))
            elif keyword == "atom":
                if len(tokens) < 1:
                    raise ValueError("keyword atom requires a list argument")
                num_list = tokens.pop(0)
                i_atoms = parse_number_list(num_list)
            elif keyword == "resi":
                if len(tokens) < 1:
                    raise ValueError("keyword resi requires a list argument")
                num_list = tokens.pop(0)
                i_atoms = select_resi(pmd, num_list)
            elif keyword == "pocket":
                i_atoms = select_residue_contacts(pmd, get_resnames("ligand"))
            elif keyword == "near":
                if len(tokens) < 1:
                    raise ValueError("keyword near requires a resname argument")
                lig_resname = tokens.pop(0)
                args = [pmd, [lig_resname]]
                if len(tokens) and tokens[0].isdigit():
                    args.append(int(tokens.pop(0)))
                i_atoms = select_residue_contacts(*args)
            elif keyword == "resname":
                if len(tokens) < 1:
                    raise ValueError("keyword resname requires an argument")
                resname = tokens.pop(0)
                i_atoms = select_resnames(pmd, [resname])
                if not len(i_atoms):
                    logger.warning(
                        f"Warning: no atoms were found for resname={resname}"
                    )
            elif keyword == "noh":
                raise ValueError(
                    f"Deprecated 'noh' in '{expr}'; use 'not {{amber @/H}}"
                )
            elif keyword == "nosolvent":
                raise ValueError(
                    f"Deprecated 'nosolvent' in '{expr}'; use 'not {{solvent}}"
                )
            elif keyword == "nosolvent":
                raise ValueError(f"Deprecated 'resid' in '{expr}'; use 'resi'")
            result.extend(i_atoms)
    return py_.uniq(result)


def get_n_residue_of_mask(pmd: parmed.Structure, mask: str):
    n_res = pmd.topology.getNumResidues()
    residues = numpy.zeros(n_res, dtype=int)
    i_atoms = select_mask(pmd, mask, is_fail_on_empty=False)
    n_atom = pmd.topology.getNumAtoms()
    if len(i_atoms) == n_atom:
        return n_res
    for a in slice_parmed(pmd, i_atoms):
        residues[a.residue.idx] = 1
    return residues.sum()


def slice_parmed_with_mask(pmd, mask):
    return slice_parmed(pmd, select_mask(pmd, mask))



import mdtraj
import mdtraj as md
import numpy
from mdtraj.utils import ilen, in_units_of


def filter_for_atom_lines(lines):
    return [l for l in lines if l.startswith("ATOM") or l.startswith("HETATM")]


def get_pdb_lines(
    mdtraj_topology,
    positions_angstroms,
    unitcell_lengths=None,
    unitcell_angles=None,
    bfactors=None,
    model_index=None,
    is_write_header=True,
):
    result = []

    if is_write_header and unitcell_lengths is not None and unitcell_angles is not None:
        box = list(unitcell_lengths) + list(unitcell_angles)
        assert len(box) == 6
        result.append(
            "CRYST1%9.3f%9.3f%9.3f%7.2f%7.2f%7.2f P 1           1 " % tuple(box)
        )

    if ilen(mdtraj_topology.atoms) != len(positions_angstroms):
        raise ValueError("The number of positions must match the number of atoms")
    if numpy.any(numpy.isnan(positions_angstroms)):
        raise ValueError("Particle position is NaN")
    if numpy.any(numpy.isinf(positions_angstroms)):
        raise ValueError("Particle position is infinite")

    if bfactors is None:
        bfactors = ["{0:5.2f}".format(0.0)] * len(positions_angstroms)
    else:
        if (numpy.max(bfactors) >= 100) or (numpy.min(bfactors) <= -10):
            raise ValueError("bfactors must be in (-10, 100)")
        bfactors = ["{0:5.2f}".format(b) for b in bfactors]

    if model_index is not None:
        result.append("MODEL     %4d" % model_index)

    i_atom = 1
    i_pos = 0
    _chain_names = [chr(ord("A") + i) for i in range(26)]
    for chainIndex, chain in enumerate(mdtraj_topology.chains):
        chainName = _chain_names[chainIndex % len(_chain_names)]
        residues = list(chain.residues)
        for resIndex, res in enumerate(residues):
            if len(res.name) > 3:
                resName = res.name[:3]
            else:
                resName = res.name
            for atom in res.atoms:
                if (
                    len(atom.name) < 4
                    and atom.name[:1].isalpha()
                    and (atom.element is None or len(atom.element.symbol) < 2)
                ):
                    atomName = " " + atom.name
                elif len(atom.name) > 4:
                    atomName = atom.name[:4]
                else:
                    atomName = atom.name
                coords = positions_angstroms[i_pos]
                if atom.element is not None:
                    symbol = atom.element.symbol
                else:
                    symbol = " "
                if atom.serial is not None and len(mdtraj_topology._chains) < 2:
                    # We can't do this for more than 1 chain
                    # to prevent issue 1611
                    atomSerial = atom.serial
                else:
                    atomSerial = i_atom
                line = (
                    "ATOM  %5d %-4s %3s %1s%4d    %s%s%s  1.00 %5s      %-4s%2s  "
                    % (  # Right-justify atom symbol
                        atomSerial % 100000,
                        atomName,
                        resName,
                        chainName,
                        (res.resSeq) % 10000,
                        _format_83(coords[0]),
                        _format_83(coords[1]),
                        _format_83(coords[2]),
                        bfactors[i_pos],
                        atom.segment_id[:4],
                        symbol[-2:],
                    )
                )
                assert len(line) == 80, "Fixed width overflow detected"
                result.append(line)
                i_pos += 1
                i_atom += 1
            if resIndex == len(residues) - 1:
                result.append(
                    "TER   %5d      %3s %s%4d"
                    % (atomSerial + 1, resName, chainName, res.resSeq)
                )
                i_atom += 1

    if model_index is not None:
        result.append("ENDMDL")

    return result


def get_pdb_lines_of_traj_frame(
    traj: mdtraj.Trajectory,
    i_frame=0,
    bfactors=None,
    model_index=None,
    is_write_header=True,
):
    """
    Adapted from mdtraj.formats.pdb.pdbfile.py
    """
    positions = in_units_of(traj._xyz[i_frame], traj._distance_unit, "angstroms")

    if traj.unitcell_lengths is None and traj.unitcell_angles is None:
        unitcell_lengths = None
        unitcell_angles = None
    elif traj.unitcell_lengths is None:
        raise ValueError("unitcell_lengths is missing")
    elif traj.unitcell_angles is None:
        raise ValueError("unitcell_angles is missing")
    else:
        unitcell_lengths = traj.unitcell_lengths[i_frame]
        unitcell_lengths = in_units_of(
            unitcell_lengths, traj._distance_unit, "angstroms"
        )
        if not len(unitcell_lengths) == 3:
            raise ValueError("unitcell_lengths must be length 3")
        unitcell_angles = traj.unitcell_angles[i_frame]
        if not len(unitcell_angles) == 3:
            raise ValueError("unitcell_angles must be length 3")

    return get_pdb_lines(
        traj.topology,
        positions,
        unitcell_lengths,
        unitcell_angles,
        bfactors,
        model_index,
        is_write_header,
    )


def save_traj_frame_pdb(traj, pdb, i_frame=0, model_index=None):
    with open(pdb, "w") as f:
        for line in get_pdb_lines_of_traj_frame(
            traj, i_frame=i_frame, model_index=model_index
        ):
            f.write(line + "\n")


def _format_83(f):
    """Format a single float into a string of width 8, with ideally 3 decimal
    places of precision. If the number is a little too large, we can
    gracefully degrade the precision by lopping off some of the decimal
    places. If it's much too large, we throw a ValueError"""
    if -999.999 < f < 9999.999:
        return "%8.3f" % f
    if -9999999 < f < 99999999:
        return ("%8.3f" % f)[:8]
    raise ValueError(
        'coordinate "%s" could not be represnted ' "in a width-8 field" % f
    )


def remove_model_lines(pdb):
    """
    Remove the "MODEL" and "ENDMDL" lines so as not confuse parmed
    """
    has_model_line = False
    lines = []
    with open(pdb) as f:
        for l in f.readlines():
            if "MODEL" in l:
                has_model_line = True
            elif "ENDMDL" in l:
                break
            else:
                lines.append(l)
    if has_model_line:
        with open(pdb, "w") as f:
            f.write("".join(lines))


def _split_pdb_file(pdb_file, ligand_residue_name) -> (str, str):
    pdb_file = str(pdb_file)
    traj = md.load_pdb(pdb_file)

    protein_ixs = traj.top.select(f"not resname {ligand_residue_name}")
    ligand_ixs = traj.top.select(f"resname {ligand_residue_name}")

    if len(ligand_ixs) == 0:
        raise ValueError(
            f"No ligand with residue name {ligand_residue_name} found in {pdb_file}"
        )

    protein_pdb_file = pdb_file.replace(".pdb", "_protein.pdb")
    protein = traj.atom_slice(protein_ixs)
    protein.save(protein_pdb_file)

    ligand_pdb_file = pdb_file.replace(".pdb", f"_{ligand_residue_name}.pdb")
    traj = traj.atom_slice(ligand_ixs)
    traj.save(ligand_pdb_file)

    return protein_pdb_file, ligand_pdb_file



import itertools
import logging
import pickle

import mdtraj
import parmed
import pydash as py_
from parmed import unit
from path import Path

from .pdb import remove_model_lines

logger = logging.getLogger(__name__)

__doc__ = """
Useful transforms for parmed.Structure, mdtraj.Trajectory, and OpenMM.
Loads structures using convenient transforms and unit conversions.

We have found that saving parmed.Structure.__getstate__() as a pickle,
is a very efficient and fast way of saving cross-platform MD prep files.

For units:
  - mdtraj and openmm use nanometers
  - pdb & parmed use angstroms

"""


def dump_parmed(pmd: parmed.Structure, fname: str):
    with open(fname, "wb") as handle:
        pickle.dump(file=handle, obj=pmd.__getstate__())


def load_parmed(fname: str) -> parmed.Structure:
    from packaging import version
    from parmed.topologyobjects import TrackedList

    with open(fname, "rb") as handle:
        state = pickle.load(file=handle)
        if version.parse(parmed.__version__) > version.parse("3.4.3"):
            if "links" not in state:
                state["links"] = TrackedList()
        pmd = parmed.structure.Structure()
        pmd.__setstate__(state)
    return pmd


def get_parmed_from_pdb(pdb: str) -> parmed.Structure:
    """
    Reads pdb with a sanity check for model lines that confuses parmed
    """
    suffix = Path(pdb).ext.lower()
    if not suffix == ".pdb":
        raise ValueError(f"Can't process {pdb} of type {suffix}, only .pdb")
    # Check for issue where mdtraj saves MODEL 0, which throws error in parmed
    remove_model_lines(pdb)
    return parmed.load_file(pdb)


def get_parmed_from_parmed_or_pdb(pdb_or_parmed: str) -> parmed.Structure:
    """
    :param pdb_or_parmed: str - either .parmed or .pdb
    """
    suffix = Path(pdb_or_parmed).ext
    if suffix == ".pdb":
        pmd = get_parmed_from_pdb(pdb_or_parmed)
    elif suffix == ".parmed":
        pmd = load_parmed(pdb_or_parmed)
    else:
        raise ValueError(f"Can't process {pdb_or_parmed} of type {suffix}")
    return pmd


def get_parmed_from_mdtraj(traj: mdtraj.Trajectory, i_frame=0) -> parmed.Structure:
    return parmed.openmm.load_topology(traj.top.to_openmm(), xyz=traj.xyz[i_frame])


def get_parmed_from_openmm(openmm_topology, positions=None) -> parmed.Structure:
    """
    :param positions: unit.Quantity(dist) | [float] in angstroms
    """
    return parmed.openmm.load_topology(openmm_topology, xyz=positions)


def get_mdtraj_from_parmed(pmd: parmed.Structure) -> mdtraj.Trajectory:
    return mdtraj.Trajectory(
        xyz=pmd.coordinates / 10, topology=mdtraj.Topology.from_openmm(pmd.topology)
    )


def get_mdtraj_from_openmm(openmm_topology, openmm_positions) -> mdtraj.Trajectory:
    if unit.is_quantity(openmm_positions):
        openmm_positions = openmm_positions.value_in_unit(unit.nanometer)
    mdtraj_topology = mdtraj.Topology.from_openmm(openmm_topology)
    return mdtraj.Trajectory(topology=mdtraj_topology, xyz=openmm_positions)


def calc_residue_contacts_with_mdtraj(
    traj, i_residues1, i_residues2, cutoff_nm=None, max_n_residue=None
) -> [int]:
    """
    :return: [int] - indices of closest residues to ligand
    """
    # Generate pairs of residue indices [[i_lig1, i_res1], [i_lig1, i_res2]....]
    pairs = list(itertools.product(i_residues1, i_residues2))

    # Calculate distances as nx1 numpy.array and pairs is nx2 numpy.array
    # periodic=False turns off period cell correction
    distances, pairs = mdtraj.compute_contacts(
        traj, contacts=pairs, scheme="closest-heavy", periodic=False
    )

    # Get sorted top_entries list of contact residues
    top_entries = [(d, pair[1]) for d, pair in zip(distances[0], pairs)]
    top_entries = py_.sort_by(top_entries, lambda e: e[0])
    if max_n_residue:
        top_entries = top_entries[:max_n_residue]
    if cutoff_nm:
        top_entries = py_.filter_(top_entries, lambda e: e[0] <= cutoff_nm)

    return [e[1] for e in top_entries]


def slice_parmed(pmd: parmed.Structure, i_atoms: [int]) -> parmed.Structure:
    # This function avoids the issue where parmed expects a bit
    # mask for selections for full selections but atom indices otherwise
    return pmd if len(i_atoms) == len(pmd.atoms) else pmd[i_atoms]



#!/usr/bin/env python
import logging

import click
import mdtraj
from path import Path

import easytrajh5.show
from easytrajh5.h5 import EasyH5File
from easytrajh5.manager import TrajectoryManager
from easytrajh5.select import get_n_residue_of_mask, select_mask
from easytrajh5.struct import slice_parmed, get_parmed_from_parmed_or_pdb
from easytrajh5.traj import EasyTrajH5File

logger = logging.getLogger(__name__)

logging.basicConfig(format="%(message)s", level=logging.INFO)


@click.group()
def h5():
    """
    h5: preprocessing and analysis tools
    """
    pass


@h5.command(no_args_is_help=True)
@click.argument("h5")
def schema(h5):
    """Examine layout of H5"""
    EasyH5File(h5).print_schema()


@h5.command(no_args_is_help=True)
@click.argument("h5")
@click.argument("dataset", default=None, required=False)
@click.option(
    "--frames", default=None, help="list of frames e.g. '1,2,3-10,11,200-211'"
)
@click.option("--json", is_flag=True, default=False, help="Format json")
def dataset(h5, dataset, frames, json):
    """Examine contents of h5"""
    f = EasyH5File(h5)
    title = h5
    if dataset is None:
        f.print_dataset_table(title)
    else:
        print(f"\n  {title}")
        f.print_dataset(dataset, frames=frames, is_json=json)


@h5.command(no_args_is_help=True)
@click.argument("h5", default="trajectory.h5")
@click.option("--mask", default=None, help="atom selection", show_default=True)
@click.option("--i", "-i", default=0, help="frame", show_default=True)
def pdb(h5_trajectory, mask, i):
    """
    Extract PDB of a frame of an H5
    """
    pdb = Path(h5_trajectory).with_suffix(".pdb")
    pmd = EasyTrajH5File(h5_trajectory, atom_mask=mask).get_topology_parmed(i_frame=i)
    pmd.save(pdb, overwrite=True)
    print(f"Wrote {pdb=} {i=} {mask=}")


@h5.command(no_args_is_help=True)
@click.argument("h5-pdb-parmed")
@click.argument("mask", default=None, required=False)
@click.option("--pdb", help="Save to PDB")
@click.option("--atom", flag_value=True, help="List all atoms")
@click.option("--res", flag_value=True, help="List all residues")
@click.option("--i", "-i", default=0, help="frame", show_default=True)
def mask(h5_pdb_parmed, mask, pdb, atom, res, i):
    """
    Explore residues/atoms of H5/PDB/PARMED using mask
    """
    filename = Path(h5_pdb_parmed)
    ext = filename.ext.lower()

    if ext not in [".h5", ".pdb", ".parmed"]:
        print(f"Can't recognize file extension {h5_pdb_parmed}")
        return

    if ext == ".h5":
        pmd = EasyTrajH5File(h5_pdb_parmed).get_topology_parmed(i_frame=i)
    else:
        pmd = get_parmed_from_parmed_or_pdb(h5_pdb_parmed)

    if mask is None:
        get_n_residue_of_mask(pmd, "protein")
        get_n_residue_of_mask(pmd, "ligand")
        get_n_residue_of_mask(pmd, "solvent")
        get_n_residue_of_mask(pmd, "not {merge {protein} {solvent} {ligand}}")
        return

    i_atoms = select_mask(pmd, mask, is_fail_on_empty=False)
    if not len(i_atoms):
        print("Could select any atoms")
        return

    pmd = slice_parmed(pmd, i_atoms)
    if res:
        residues = []
        for a in pmd.atoms:
            if a.residue not in residues:
                residues.append(a.residue)
        for residue in residues:
            print(residue)
    if atom:
        for a in pmd.atoms:
            print(a)

    if not pdb:
        return

    pdb = Path(pdb).with_suffix(".pdb")
    pmd.save(pdb, overwrite=True)
    print(f"Wrote {pdb}")


@h5.command(no_args_is_help=True)
@click.argument("h5_list", nargs=-1)
@click.option(
    "--prefix",
    default="merged",
    help="prefix for newly generated .h5",
    show_default=True,
)
@click.option(
    "--mask",
    default="amber @*",
    help="selection mask to specify atoms in newly generated .h5",
    show_default=True,
)
def merge(h5_list, prefix, mask):
    """Merge a list of H5 files"""
    traj_mananger = TrajectoryManager(paths=h5_list, atom_mask=mask)
    frames = []
    for i_traj in traj_mananger.traj_file_by_i.keys():
        for i_frame in range(0, traj_mananger.get_n_frame(i_traj)):
            frames.append(traj_mananger.read_as_frame_traj((i_frame, i_traj)))
    full_traj = mdtraj.join(frames)
    full_traj_h5 = Path(prefix).with_suffix(".h5")
    print(f"Merged {h5_list} --> {full_traj_h5}")
    full_traj.save_hdf5(full_traj_h5)


@h5.command(no_args_is_help=True)
@click.argument("filename")
@click.argument("mask", required=False)
@click.argument("mask2", required=False)
@click.option("-s", "--sphere", is_flag=True, help="Show all atoms as spheres")
@click.option("-t", "--stick", is_flag=True, help="Show all bonds as stick")
@click.option(
    "-k",
    "--keep-solvent",
    is_flag=True,
    default=False,
    show_default=True,
    help="Keep solvent",
)
def show_pymol(filename, mask, mask2, sphere, stick, keep_solvent):
    """
    Use PYMOL to show H5/PDB/PARMED with mask

    \b
    Default: no solvent
    Optional two masks (green/purple). Atom mask selections, see:
      https://github.com/redesignScience/easytrajh5#atom-selection-language
    """
    easytrajh5.show.pymol(**locals())


@h5.command(no_args_is_help=True)
@click.argument("filename")
@click.argument("parmed")
@click.argument("mask", required=False)
@click.argument("mask2", required=False)
@click.option("-s", "--sphere", is_flag=True, help="Show all atoms as spheres")
@click.option("-t", "--stick", is_flag=True, help="Show all atoms as sticks")
@click.option(
    "-k",
    "--keep-solvent",
    is_flag=True,
    default=False,
    show_default=True,
    help="Keep solvent",
)
def show_chimera(filename, parmed, mask, mask2, sphere, stick, keep_solvent):
    """
    Use CHIMERA to show H5/PDB/PARMED with mask, needs PARMED

    \b
    Default: no solvent
    Optional two masks (green/purple). Atom mask selections, see:
      https://github.com/redesignScience/easytrajh5#atom-selection-language
    """
    easytrajh5.show.chimera(**locals())


@h5.command(no_args_is_help=True)
@click.argument("filename")
@click.argument("mask", required=False)
@click.argument("mask2", required=False)
@click.option("-s", "--sphere", is_flag=True, help="Show all atoms as spheres")
@click.option(
    "-k",
    "--keep-solvent",
    is_flag=True,
    default=False,
    show_default=True,
    help="Keep solvent",
)
def show_vmd(filename, mask, mask2, sphere, keep_solvent):
    """
    Use VMD to show H5/PDB/PARMED with mask

    \b
    Default: no solvent
    Optional two masks (green/purple). Atom mask selections, see:
      https://github.com/redesignScience/easytrajh5#atom-selection-language
    """
    easytrajh5.show.vmd(**locals())


if __name__ == "__main__":
    h5()



import logging
import operator
import pickle
from collections import namedtuple
from typing import TypeVar, Sequence

import numpy
import parmed
from addict import Dict
from mdtraj import Topology, Trajectory
from mdtraj.core import element as elem
from mdtraj.reporters.basereporter import _BaseReporter
from mdtraj.utils import ensure_type, in_units_of
from mdtraj.utils.unitcell import lengths_and_angles_to_box_vectors
from path import Path

from easytrajh5.fs import load_yaml, dump_yaml
from easytrajh5.pdb import remove_model_lines
from .fs import tic, toc
from .h5 import EasyH5File
from .select import select_mask
from .struct import get_parmed_from_openmm, slice_parmed

logger = logging.getLogger(__name__)

_Slice = TypeVar("_Slice", bound=Sequence)


class EasyTrajH5File(EasyH5File):
    """
    EasyTrajH5File is an object used to conviently process mdtraj-type
    H5 files efficiently, using the internal indexing features
    of H5 via the h5py library.

    In particular, we provide a convenient atom selection language
    to pre-select for atoms before extracting the trajectory. This
    is used to initialize the object via the topology, and subsequent
    read methods will read only the selected atoms.

    As well, a special dry_cache option has been implemented that
    will store the dry topology ("not solvent") in a special cache
    inside the H5 file.  Accessing the reduced dry topology will be
    fast, as the H5 can skip the loading of the full topology.
    Subsequent frame reads will only reference the dry atoms.

    The trajectory can be returned as:

        - read_as_traj() - the whole trajectory
        - read_frame_as_traj() - a traj with one frame
        - read_frame_slice_as_traj() - a selection of frames

    As well, the API of mdtraj.formats.H5TrajectoryFile has been replicated,
    and EasyTrajH5File can serve as a backend of openmm._BaseReporter:

        - __init__(h5, mode)
        - @property.setter topology
        - distance_unit
        - flush (optional)
        - close
        - read
        - write
    """

    # Compatibility with mdtraj.formats.hdf5.H5TrajectoryFile.read
    # nanometers
    distance_unit = Trajectory._distance_unit

    fields = [
        Dict(
            key="coordinates",
            shape=(None, None, 3),
            units=distance_unit,
            traj_key="xyz",
        ),
        Dict(key="time", shape=(None,), units="picoseconds", traj_key="time"),
        Dict(
            key="cell_lengths",
            shape=(None, 3),
            units=distance_unit,
            traj_key="unitcell_lengths",
        ),
        Dict(
            key="cell_angles",
            shape=(None, 3),
            units="degrees",
            traj_key="unitcell_angles",
        ),
        Dict(key="velocities", shape=(None, None, 3), units="nanometers/picosecond"),
        Dict(key="kineticEnergy", shape=(None,), units="kilojoules_per_mole"),
        Dict(key="potentialEnergy", shape=(None,), units="kilojoules_per_mole"),
        Dict(key="temperature", shape=(None,), units="kelvin"),
        # reversed the mdtraj choice to rename this `lambda` in the h5
        Dict(key="alchemicalLambda", shape=(None,), units="dimensionless"),
    ]

    default_attrs = {
        "conventions": "Pande",
        "conventionVersion": "1.1",
        "program": "rseed",
        "programVersion": "0.1",
        "application": "rseed",
    }

    def __init__(
        self,
        fname: str,
        mode: str = "a",
        atom_mask: str = "",
        is_dry_cache: bool = False,
    ):
        logger.info(f"{self.__class__.__name__}: {fname=} {mode=} {atom_mask=} {is_dry_cache=}")
        logger.info(tic("open connection"))
        super().__init__(fname, mode)
        logger.info(toc())

        # to reach here, we've successfully loaded an .h5 file
        if mode == "w":
            self.has_header = False
        elif mode in ["a", "r"]:
            self.has_header = self.has_dataset("coordinates")
        else:
            raise ValueError("mode must be one of ['r', 'w', 'a']")

        # stores the full topology
        self._topology = None
        self.topology_by_atom_hash = {}

        self.atom_mask = atom_mask
        self.atom_indices = None
        self.last_i_frame = None
        self.last_frame = None
        self.last_atom_indices = None

        if is_dry_cache:
            if atom_mask and atom_mask != "not {solvent}":
                raise ValueError(
                    "Can't set is_dry_cache=True and atom_mask at the same time"
                )
            atom_mask = "not {solvent}"
            logger.info(f"{mode=} {atom_mask=}")
            if mode == "r":
                logger.info("Warning: can't use dry_atom_cache in read-only mode")
            else:
                dry_atom_indices = self.load_dry_topology_from_cache()
                self.atom_indices = dry_atom_indices
                return

        if atom_mask:
            logger.info("setting atom_indices from atom_mask")
            sliced_top, atom_indices = self.slice_topology(atom_mask)
            self.atom_indices = atom_indices
            atom_hash = tuple(atom_indices)
            self.topology_by_atom_hash[atom_hash] = sliced_top

    def fetch_topology(self, atom_indices=None) -> Topology:
        if atom_indices is not None:
            atom_hash = tuple(atom_indices)
            if atom_hash in self.topology_by_atom_hash:
                return self.topology_by_atom_hash[atom_hash]

        # Need to check if self._topology exists
        if self._topology is None:
            if not self.has_dataset("topology"):
                raise ValueError(f"No topology saved in {self.fname}")
            logger.info(tic("loading topology"))
            topology_dict = self.get_json_dataset("topology")
            self._topology = get_mdtraj_topology_from_dict(topology_dict)
            logger.info(toc())

        if atom_indices is not None:
            atom_hash = tuple(atom_indices)
            logger.info(tic(f"slicing topology {len(atom_indices)}"))
            topology = self._topology.subset(atom_indices)
            logger.info(toc())
            self.topology_by_atom_hash[atom_hash] = topology
            return topology
        else:
            return self._topology

    def load_dry_topology_from_cache(self):
        if self.has_dataset("dry_atoms"):
            logger.info(tic("reading dry_atoms"))
            dry_atom_indices = self.get_dataset("dry_atoms")[:]
            logger.info(toc())

            if len(dry_atom_indices) == 0:
                logger.info("dry_atoms=[] -> no dry_topology")
                dry_atom_indices = None
            elif self.has_dataset("dry_topology"):
                logger.info(tic("reading dry_topology"))
                topology_dict = self.get_json_dataset("dry_topology")
                dry_top = get_mdtraj_topology_from_dict(topology_dict)
                logger.info(toc())
                atom_hash = tuple(dry_atom_indices)
                self.topology_by_atom_hash[atom_hash] = dry_top
        else:
            # haven't tried calculating dry_topology yet
            dry_top, dry_atom_indices = self.slice_topology("not {solvent}")
            n_atom = sum(1 for _ in self.topology.atoms)
            if len(dry_atom_indices) == n_atom:
                logger.info(tic("no solvent => save dry_atoms=[]"))
                self.set_array_dataset("dry_atoms", numpy.array([]))
                logger.info(toc())
                dry_atom_indices = None
            else:
                logger.info(tic("saving dry_topology and dry_atoms"))
                self.set_json_dataset(
                    "dry_topology", get_dict_from_mdtraj_topology(dry_top)
                )
                self.set_array_dataset("dry_atoms", numpy.array(dry_atom_indices))
                self.flush()
                logger.info(toc())
                atom_hash = tuple(dry_atom_indices)
                self.topology_by_atom_hash[atom_hash] = dry_top

        return dry_atom_indices

    def get_topology_parmed(self, i_frame=0):
        positions_angstroms = self.get_dataset("coordinates")[i_frame] * 10
        if self.atom_indices is None:
            mdtraj_topology = self.topology
        else:
            logger.info("Select atom_mask")
            mdtraj_topology = self.fetch_topology(self.atom_indices)
            positions_angstroms = positions_angstroms[self.atom_indices]
        return get_parmed_from_openmm(mdtraj_topology.to_openmm(), positions_angstroms)

    def get_parmed_from_dataset(self, i_frame=None):
        if not self.has_dataset("parmed"):
            return None

        pmd = parmed.Structure()
        blob = self.get_bytes_dataset("parmed")
        pmd.__setstate__(pickle.loads(blob))

        if i_frame is not None:
            coordinates = self.get_dataset("coordinates")
            if i_frame < 0:
                i_frame = coordinates.shape[0] + i_frame
            pmd.positions = coordinates[i_frame] * 10.0
            box_vectors = self.read_parmed_box_vectors(i_frame)
            if box_vectors is not None:
                pmd.box_vectors = box_vectors

        return pmd

    def slice_topology(self, atom_mask: str) -> (Topology, [int]):
        logger.info(tic("building parmed"))
        positions_angstroms = self.get_dataset("coordinates")[0] * 10
        pmd = get_parmed_from_openmm(self.topology.to_openmm(), positions_angstroms)
        logger.info(toc())

        logger.info(tic("select mask"))
        i_atoms = select_mask(pmd, atom_mask, is_fail_on_empty=False)
        logger.info(toc())

        logger.info(tic("slicing parmed"))
        sliced_pmd = slice_parmed(pmd, i_atoms)
        logger.info(toc())

        logger.info(tic("converting to mdtraj topology"))
        sliced_top = Topology.from_openmm(sliced_pmd.topology)
        logger.info(toc())

        return sliced_top, i_atoms

    @property
    def topology(self) -> Topology:
        if self._topology is None:
            self.fetch_topology()
        return self._topology

    @topology.setter
    def topology(self, topology):
        """Compatibility with mdtraj.formats.hdf5.H5TrajectoryFile.topology="""
        self._topology = topology
        self.set_json_dataset("topology", get_dict_from_mdtraj_topology(topology))
        self.handle.flush()

    def get_n_frame(self):
        return self.get_dataset("coordinates").shape[0]

    def __len__(self):
        return self.get_n_frame()

    def read_atom_dataset_progressively(
        self, key, frame_slice, atom_indices=slice(None)
    ) -> numpy.ndarray:
        """Returns numpy.ndarrary[float] of [n_frame x <shape of frame of dataset>]"""

        stride = frame_slice.step or 1
        start = frame_slice.start or 0

        dataset = self.get_dataset(key)
        n_frame_total = frame_slice.stop - start
        n_atom = (
            len(atom_indices) if isinstance(atom_indices, list) else dataset.shape[1]
        )
        n_frame_in_chunk = dataset.chunks[0]
        n_frame = n_frame_total // stride
        offset = 1 if n_frame % stride != 0 else 0
        result = numpy.empty((n_frame + offset, n_atom, 3), dataset.dtype)
        # Try this n_frame_in_page first, and let errors force us to reduce it
        n_frame_of_page = n_frame_total

        i_frame_of_page = start
        i_frame_in_result = 0
        while True:
            try:
                frame_slice = slice(
                    i_frame_of_page, i_frame_of_page + n_frame_of_page, stride
                )
                data = dataset[frame_slice, atom_indices, :]
            except IOError as ioe:
                if ioe.errno == 413 and n_frame_of_page > n_frame_in_chunk:
                    n_frame_of_page = max(n_frame_in_chunk, n_frame_of_page // 2)
                    continue
                else:
                    raise IOError(f"Error retrieving data: {ioe.errno}")

            n_frame_in_data = data.shape[0]
            frame_slice = slice(i_frame_in_result, i_frame_in_result + n_frame_in_data)
            result[frame_slice, slice(None), slice(None)] = data

            i_frame_in_result += n_frame_in_data
            i_frame_of_page += n_frame_of_page

            # make sure page starts on a stride step
            offset = i_frame_of_page % stride
            if offset != 0:
                i_frame_of_page += stride - offset

            if i_frame_of_page >= frame_slice.stop:
                break

        return result

    def read(self, n_frames=None, stride=None, atom_indices=None) -> namedtuple:
        """
        Compatibility with mdtraj.formats.hdf5.H5TrajectoryFile.read

        :returns namedtuple:
             The returned namedtuple will have the fields "coordinates", "time", "cell_lengths",
             "cell_angles", "velocities", "kineticEnergy", "potentialEnergy",
             "temperature" and "alchemicalLambda". Each of the fields in the
             returned namedtuple will either be a numpy array or None, dependening
             on if that data was saved in the trajectory. All of the data shall be
             n units of "nanometers", "picoseconds", "kelvin", "degrees" and
             "kilojoules_per_mole".
        """
        if n_frames is None:
            n_frames = numpy.inf
        if stride is not None:
            stride = int(stride)

        total_n_frames = self.get_n_frame()
        frame_slice = slice(0, min(n_frames, total_n_frames), stride)
        if frame_slice.stop - frame_slice.start == 0:
            return []

        if atom_indices is None:
            # get all the atoms
            atom_indices = slice(None)
        # TODO: check atom_indices for type and fit to n_frames

        keys = [field["key"] for field in self.fields]
        Frames = namedtuple("Frames", keys)
        kwargs = {}
        for key in keys:
            if self.has_dataset(key):
                kwargs[key] = self.read_atom_dataset_progressively(
                    key, frame_slice, atom_indices
                )
            else:
                kwargs[key] = None
        return Frames(**kwargs)

    def read_frame_slice_as_traj(
        self, frame_slice, atom_indices, coordinates_only=False
    ):
        """
        :param frame_slice: int | slice(0, n, stride)
        :param atom_indices: list[int] | None
        """

        kwargs = Dict(topology=self.fetch_topology(atom_indices))

        if atom_indices is None:
            atom_indices = slice(None)

        logger.info(tic("reading frames"))
        if coordinates_only:
            fields = [field for field in self.fields if field["key"] == "coordinates"]
        else:
            fields = self.fields

        for field in fields:
            if "traj_key" in field and self.has_dataset(field.key):
                dataset = self.get_dataset(field.key)

                if len(field.shape) == 3:
                    if isinstance(frame_slice, int):
                        data = dataset[frame_slice, atom_indices, :]
                    else:
                        data = self.read_atom_dataset_progressively(
                            field.key, frame_slice, atom_indices
                        )
                elif len(field.shape) == 2:
                    data = dataset[frame_slice, :]
                else:
                    data = dataset[frame_slice]

                kwargs[field.traj_key] = data
        logger.info(toc())

        traj = Trajectory(**kwargs)

        return traj

    def read_as_traj(self, stride=1, atom_indices=None):
        if atom_indices is None:
            atom_indices = self.atom_indices
        n = self.get_n_frame()
        # NOTE: slice can only start on 0
        return self.read_frame_slice_as_traj(slice(0, n, stride), atom_indices)

    def read_frame_as_traj(self, i_frame, atom_indices=None):
        if atom_indices is None:
            atom_indices = self.atom_indices

        if i_frame < 0:
            i_frame = self.get_n_frame() + i_frame

        if self.last_i_frame == i_frame:
            if numpy.array_equal(self.last_atom_indices, atom_indices):
                logger.info(f"same as last frame {i_frame}")
                return self.last_frame

        frame = self.read_frame_slice_as_traj(i_frame, atom_indices)

        self.last_frame = frame
        self.last_i_frame = i_frame
        self.last_atom_indices = atom_indices

        return frame

    def read_parmed_box_vectors(self, i_frame):
        """
        :return: iterable of 3, each 3-element tuple of unit cell vectors in angstroms

        """
        if not self.has_dataset("cell_lengths") or not self.has_dataset("cell_angles"):
            return None
        these_cell_lengths = self.get_dataset("cell_lengths")[i_frame]
        these_cell_angles = self.get_dataset("cell_angles")[i_frame]
        v1, v2, v3 = lengths_and_angles_to_box_vectors(
            these_cell_lengths[0],  # a
            these_cell_lengths[1],  # b
            these_cell_lengths[2],  # c
            these_cell_angles[0],  # alpha
            these_cell_angles[1],  # beta
            these_cell_angles[2],  # gamma
        )
        unitcell_vectors = numpy.swapaxes(numpy.dstack((v1, v2, v3)), 1, 2) * 10
        return unitcell_vectors[0]

    def write_header(self, n_atoms, keys):
        for k, v in self.default_attrs.items():
            self.set_attr(k, v)

        for field in self.fields:
            if field.key not in keys:
                continue

            frame_shape = tuple(field.shape[1:])
            if len(frame_shape) > 1 and frame_shape[0] is None:
                frame_shape = (n_atoms, *frame_shape[1:])

            self.create_extendable_dataset(field.key, frame_shape, numpy.float32)
            self.set_attr("units", field.units, dataset_key=field.key)

    def write(
        self,
        coordinates,
        time=None,
        cell_lengths=None,
        cell_angles=None,
        velocities=None,
        kineticEnergy=None,
        potentialEnergy=None,
        temperature=None,
        alchemicalLambda=None,
    ):
        """
        Compatibility with mdtraj.formats.hdf5.H5TrajectoryFile.write
        """
        logger.info(tic("writing coordinates"))
        frames_by_key = {}
        for field in self.fields:
            # trick to read the args by their string name
            frames = locals().get(field.key)
            if frames is not None:
                frames_by_key[field.key] = ensure_type(
                    in_units_of(frames, None, field.units),
                    name=field.key,
                    dtype=numpy.float32,
                    shape=field.shape,
                    ndim=len(field.shape),
                    add_newaxis_on_deficient_ndim=True,
                    warn_on_cast=False,
                )

        if not self.has_header:
            n_atoms = frames_by_key["coordinates"].shape[-2]
            self.write_header(n_atoms, list(frames_by_key.keys()))
            self.has_header = True

        for key, frames in frames_by_key.items():
            self.extend_dataset(key, frames)

        self.flush()
        logger.info(toc())
        logger.info(f'final shape: {frames_by_key["coordinates"].shape}')


class EasyTrajH5Reporter(_BaseReporter):
    @property
    def backend(self):
        return EasyTrajH5File


def get_dict_from_mdtraj_topology(topology):
    try:
        topology_dict = {"chains": [], "bonds": []}

        for chain in topology.chains:
            chain_dict = {"residues": [], "index": int(chain.index)}
            for residue in chain.residues:
                residue_dict = {
                    "index": int(residue.index),
                    "name": str(residue.name),
                    "atoms": [],
                    "resSeq": int(residue.resSeq),
                    "segmentID": str(residue.segment_id),
                }

                for atom in residue.atoms:
                    try:
                        element_symbol_string = str(atom.element.symbol)
                    except AttributeError:
                        element_symbol_string = ""

                    residue_dict["atoms"].append(
                        {
                            "index": int(atom.index),
                            "name": str(atom.name),
                            "element": element_symbol_string,
                        }
                    )
                chain_dict["residues"].append(residue_dict)
            topology_dict["chains"].append(chain_dict)

        for atom1, atom2 in topology.bonds:
            topology_dict["bonds"].append([int(atom1.index), int(atom2.index)])

        return topology_dict

    except AttributeError as e:
        raise AttributeError(
            "topology_object fails to implement the"
            "chains() -> residue() -> atoms() and bond() protocol. "
            "Specifically, we encountered the following %s" % e
        )


def get_mdtraj_topology_from_dict(topology_dict) -> Topology:
    topology = Topology()

    for chain_dict in sorted(topology_dict["chains"], key=operator.itemgetter("index")):
        chain = topology.add_chain()
        for residue_dict in sorted(
            chain_dict["residues"], key=operator.itemgetter("index")
        ):
            try:
                ref_seq = residue_dict["resSeq"]
            except KeyError:
                ref_seq = None
                logger.warning(
                    "No resSeq information found in HDF h5, defaulting to zero-based indices"
                )
            try:
                segment_id = residue_dict["segmentID"]
            except KeyError:
                segment_id = ""
            residue = topology.add_residue(
                residue_dict["name"], chain, resSeq=ref_seq, segment_id=segment_id
            )
            for atom_dict in sorted(
                residue_dict["atoms"], key=operator.itemgetter("index")
            ):
                try:
                    element = elem.get_by_symbol(atom_dict["element"])
                except KeyError:
                    element = elem.virtual
                topology.add_atom(atom_dict["name"], element, residue)

    atoms = list(topology.atoms)
    for index1, index2 in topology_dict["bonds"]:
        topology.add_bond(atoms[index1], atoms[index2])

    return topology


def convert_h5_to_dcd_and_pdb(h5_fname, is_solvent=True):
    dcd = Path(h5_fname).with_suffix(".dcd")
    pdb = Path(h5_fname).with_suffix(".pdb")
    yaml = Path(h5_fname).with_suffix(".rselect.yaml")

    is_convert = False
    meta = dict(is_solvent=is_solvent)
    if yaml.exists():
        old_meta = load_yaml(yaml)
        if meta != old_meta:
            is_convert = True
    if not dcd.exists():
        is_convert = True

    if is_convert:
        logger.info(f"Loading {h5_fname}")

        mask = "not {solvent}" if not is_solvent else ""
        traj = EasyTrajH5File(h5_fname, atom_mask=mask).read_as_traj()

        try:
            logger.info("Imaging frames")
            traj.image_molecules(inplace=True)
        except Exception:
            logger.error("Failed to image molecule")

        try:
            logger.info("Aligning frames")
            traj = traj.superpose(traj[0], atom_indices=traj.top.select("protein"))
        except Exception:
            logger.error("Failed to align")

        logger.info("Centering")
        traj.center_coordinates()

        logger.info(f"Saving {dcd} {pdb}")
        traj.save_dcd(dcd)

        traj[0].save_pdb(pdb)
        remove_model_lines(pdb)
        dump_yaml(meta, yaml)

    return dcd, pdb



import numpy as np
from parmed import unit
from parmed.unit import *


def compile_unit(unit_repr):
    """
    This is a hack to allow the generation of a unit from a string representation.
     It uses eval() to reconstruct the unit, and so requires all available units
     in the immediate scope. This has been pushed into this module to avoid the
     clobbering of the namespace from the unit module, and to isolate the
     usage of eval
    """
    return eval(unit_repr)


def get_dict_from_quantity(o):
    value = o._value
    if isinstance(value, np.ndarray):
        value = value.tolist()
    elif isinstance(value, list):
        value = np.array(value).tolist()
    result = {"type": "quantity", "value": value, "unit": str(o.unit)}
    if not hasattr(unit, result["unit"]):
        result["unit_repr"] = repr(o.unit)
    return result


def get_quantity_from_dict(o):
    if hasattr(unit, o["unit"]):
        return o["value"] * getattr(unit, o["unit"])
    elif o.get("unit_repr"):
        return o["value"] * compile_unit(o["unit_repr"])
    return o
